import yaml
from typing import List, Literal, Optional, TypedDict
from copy import deepcopy
import schedule
import time
from trawler.sql.extract import extract_sql

JobType = Literal['sql']


class Schedule(TypedDict):
  every: str
  interval: int = 1
  at: Optional[str]

class Job(TypedDict):
  schedule: Schedule
  type: JobType

class SQLJob(TypedDict):
  uri: str
  override_dbname: Optional[str]

class Config(TypedDict):
  jobs: List[Job]


def read_config(fname: str) -> Config:
  with open(fname) as f:
    config = yaml.load(f, Loader=yaml.FullLoader)

    return config


def run_scheduler(fname: str, now: bool):
  config = read_config(fname)

  run_on_startup = []

  for job in config["jobs"]:
    scheduled_job = getattr(
      schedule.every(job["schedule"]["interval"]),
      job["schedule"]["every"]
    )
    if job["schedule"].get("at"):
      scheduled_job.at(job["schedule"]["at"])

    if job["type"] == "sql":
      args = deepcopy(job)
      args.pop("schedule")
      args.pop("type")
      scheduled_job.do(lambda: extract_sql(**args))

    if job["schedule"].get("on_startup", False):
      run_on_startup.append(scheduled_job)

  print(f"Scheduled {len(schedule.get_jobs())} jobs")

  if now:
    schedule.run_all()
  else:
    for job in run_on_startup:
      job.run()

    while True:
      schedule.run_pending()
      time.sleep(1)



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional
from sqlalchemy.sql.sqltypes import TypeEngine

class FieldType(Enum):
    INT = "int"
    FLOAT = "float"
    STRING = "str"


@dataclass
class Field:
    name: str
    field_type: TypeEngine
    nullable: bool
    is_primary_key: bool
    comment: Optional[str]

@dataclass
class Relation:
    name: str
    source_fields: List[str]
    target_object: str
    target_fields: List[str]

@dataclass
class Table:
    name: str
    fields: List[Field]
    relations: List[Relation]
    comment: Optional[str]


@dataclass
class Schema:
    source: str




from inspect import getframeinfo, stack

from trawler.graph import Graph



# class GrpcAuth(grpc.AuthMetadataPlugin):
#     def __init__(self, client_id, client_key):
#         self._client_id = client_id
#         self._client_key = client_key

#     def __call__(self, context, callback):
#         callback((('client-id', self._client_id), ('client-key', self._client_key)), None)

# def configure(client_id, client_key):
#     global auth
#     auth = GrpcAuth(client_id, client_key)

# def record(df, locator):
#     ssl_creds = grpc.ssl_channel_credentials(root_certificates=open("../yak-collector/src/main/resources/certificate.crt", "rb").read())
#     call_credentials = grpc.metadata_call_credentials(auth)
#     channel_creds = grpc.composite_channel_credentials(ssl_creds, call_credentials)
#     channel = grpc.secure_channel('localhost:8080', channel_creds)
#     stub = trawler_grpc.TrawlerStub(channel)

#     request = trawler_proto.RecordRequest()
#     request.locator = locator
#     request.timestamp.GetCurrentTime()

#     caller = getframeinfo(stack()[1][0])
#     request.schema.comment = "%s:%d" % (caller.filename, caller.lineno)

#     for column in df.columns:
#         field = trawler_proto.Field()
#         field.name = column
#         field.type = str(df[column].dtype)
#         request.schema.field.append(field)

#         null_count = trawler_proto.Fact()
#         null_count.name = "yak.core.null-count"
#         null_count.field = field.name
#         null_count.double_value = df[column].isna().sum()
#         request.facts.facts.append(null_count)

#     row_count = trawler_proto.Fact()
#     row_count.name = "yak.core.row-count"
#     row_count.double_value = len(df)
#     request.facts.facts.append(row_count)

#     stub.Record(request)


from sqlalchemy.sql.sqltypes import TypeEngine
from sqlalchemy.dialects.postgresql import JSON, JSONB
from datetime import datetime, date
from decimal import Decimal
import logging
import sqlalchemy
import itertools

LOG = logging.getLogger(__name__)

def sql_query_scalar(conn, sql):
    LOG.debug(f"running SQL: {sql}")
    try:
        val = next(conn.execute(sql))[0]
        if isinstance(val, datetime):
            return str(val)
        elif isinstance(val, date):
            return str(val)
        if isinstance(val, Decimal):
            return str(val)
        else:
            return val
    except:
        return None

def histogram_value(conn, table_name, field_name, buckets=20):
    min_max = conn.execute(f"""
        SELECT MIN({field_name}) as min_val, MAX({field_name}) AS max_val
        FROM {table_name}
    """)
    row = next(min_max)
    min_val = row[0]
    max_val = row[1]

    if min_val == max_val:
        return None

    res = conn.execute(
    f"""SELECT 
        WIDTH_BUCKET({field_name}, {min_val}, {max_val}, {buckets}) AS buckets, COUNT(*)
        FROM {table_name}
        GROUP BY buckets
        ORDER BY buckets
    """)

    bucket_width = (max_val - min_val) / buckets

    counts = {row[0]: row[1] for row in res}

    return {
        "min": min_val,
        "max": max_val,
        "buckets": buckets,
        "counts": [
            counts.get(idx, 0)
            for idx in itertools.chain([None], range(buckets + 1))
        ]
    }

def sql_query_pairs(conn, sql):
    return [
        {"key": row[0], "value": row[1]}
        for row in conn.execute(sql)
    ]

def is_primitive_type(field_type: TypeEngine):
    try:
        return field_type.python_type in {int, str, bool, float}
    except NotImplementedError:
        pass

def is_numeric(field_type: TypeEngine):
    try:
        return field_type.python_type in {int, float}
    except NotImplementedError:
        pass

def is_date(field_type: TypeEngine):
    try:
        return field_type.python_type in {datetime, date}
    except NotImplementedError:
        pass

def get_column_metrics(engine, field_type, table_name, column_name):
    with engine.connect() as conn:
        out = {
            "metrics__nullRatio": sql_query_scalar(
                conn,
                f"""
                SELECT
                CAST(SUM(CASE WHEN {column_name} IS NULL THEN 1 ELSE 0 END) as DOUBLE PRECISION) / COUNT(*)
                FROM {table_name};
                """,
            ),
        }

        if is_primitive_type(field_type):
            out["metrics__uniqueRatio"] = sql_query_scalar(
                conn,
                f"""
                SELECT
                CAST(COUNT(DISTINCT({column_name})) as DOUBLE PRECISION) / COUNT(*)
                FROM {table_name};
                """,
            )

        if is_numeric(field_type) or is_date(field_type):
            out["metrics__max"] = sql_query_scalar(
                conn,
                f"""
                SELECT
                MAX({column_name})
                FROM {table_name};
                """,
            )

            out["metrics__min"] = sql_query_scalar(
                conn,
                f"""
                SELECT
                MIN({column_name})
                FROM {table_name};
                """,
            )

        if is_numeric(field_type):
            hist = histogram_value(
                conn,
                table_name,
                column_name
            )

            if hist:
                out["metrics__histogram"] = hist

        if field_type == JSON:
            out["metrics__countByType"] = sql_query_pairs(
                f"""
                SELECT json_typeof({column_name}), COUNT(*)
                FROM {table_name}
                GROUP BY jsonb_typeof({column_name})
                """
            )
        elif field_type == JSONB:
            out["metrics__countByType"] = sql_query_pairs(
                f"""
                SELECT jsonb_typeof({column_name}), COUNT(*)
                FROM {table_name}
                GROUP BY jsonb_typeof({column_name})
                """
            )

        return out


def get_table_metrics(engine, table_name):
    with engine.connect() as conn:
        return {
            "metrics__count": sql_query_scalar(
                conn,
                f"""
                SELECT COUNT(*)
                FROM {table_name};
                """
            )
        }



class Namespace:
    def __init__(self, base):
        self._base = base

    def entity(self, uri, **kwargs):
        pass

def namespace(base):
    return Namespace(base)

CORE = "http://trawler.dev/schema/core#"

def entity(uri, **kwargs):
    pass

def facet(uri, **kwargs):
    pass

facet("http://schema.org/name")

postgres_connector = facet(
    uri="http://trawler.dev/schema/core#postgresConnector",
    meta="json",
    json_schema={
        "type": "object"
    }
)

database_type = facet(
    uri="http://trawler.dev/schema/core#databaseType",
)



database_type = facet(
    uri="http://trawler.dev/schema/core#databaseType",
)

hostname = facet(
    uri="http://trawler.dev/schema/core#databaseType",
)

Field = entity(
    uri = CORE + "Field",
    urn = "urn:table:${type}"
)

Table = entity(
    "http://trawler.dev/schema/core#Table",
    base="table",
    schemes={
        "sql": {
            "format": "/$type/$host/$database/$table",
            "facetConstructors": {
               postgres_connector: {
                   "type": "$type",
                   "host": "$host",
                   "database": "$database",
                   "table": "$table"
               },
               hostname: "$host"
            }
        },
        "dataframe": "/${job}/${name}"
    },
)

# Are these really subtypes?
# What do we want to do e.g. about environments, you might want the environment to appear in your URN and be automatically put onto a facet.

# urn:63255f7a-e383-457a-9c30-4c7f95308749:table:sql:/postgres/db.example.com/mydb/table1
# urn::table:sql:/postgres/db.example.co/mydb/table1

# urn:<project>:<base>:<scheme>:<identifier>

has = facet(
    uri="http://trawler.dev/schema/core#has",
    meta="relationship"
)





from functools import partial
from dataclasses import dataclass
from typing import Any
import os

import requests

def to_json(value):
    if isinstance(value, EntityStub):
        return value.json()
    elif isinstance(value, list):
        return [to_json(v) for v in value]
    elif isinstance(value, dict):
        return {k: to_json(v) for k, v in value.items()}
    else:
        return value


@dataclass
class EntityStub:
    id: str
    type: str
    attributes: dict[str, Any]

    def __init__(self, typ, id, **kwargs):
        self.id = id
        self.type = typ
        self.attributes = {
            key.replace("__", ":"): value for key, value in kwargs.items()
        }

    def json(self):
        out = {"@type": self.type, "@id": self.id}

        for uri, value in self.attributes.items():
            out[uri] = to_json(value)
        return out


class Graph:
    def __init__(self):
        self._entities = []

    def add(self, entity):
        self._entities.append(entity)

    def entities(self):
        return self._entities

    def __getattr__(self, name):
        return partial(EntityStub, f"tr:{name}")

    def json(self, context=None):
        return {
            "@context": "http://trawler.dev/schema/core",
            "@graph": [entity.json() for entity in self._entities],
        }

    def store(self, project=None):
        out = self.json()

        token = os.environ["TRAWLER_TOKEN"]
        endpoint = os.environ.get("TRAWLER_ENDPOINT", "https://api.trawler.dev")
        project = project or os.environ.get("TRAWLER_PROJECT", "63255f7a-e383-457a-9c30-4c7f95308749")

        r = requests.post(
            f"{endpoint}/api/collect/{project}",
            json=out,
            headers={"X-API-Key": token}
        )
        r.raise_for_status()
        return r.json()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['trawler', 'trawler.agent', 'trawler.sql']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'click>=8.0.1,<9.0.0',
 'pandas>=1.3.3,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'schedule>=1.1.0,<2.0.0']

extras_require = \
{'sql': ['SQLAlchemy>=1.4.17,<2.0.0', 'psycopg2-binary>=2.9.1,<3.0.0']}

entry_points = \
{'console_scripts': ['trawler = trawler.agent.main:main']}

setup_kwargs = {
    'name': 'trawler-python',
    'version': '0.0.2a3',
    'description': 'Trawler is an open source metadata platform for mapping and monitoring your data',
    'long_description': '# trawler python tools\nThese are the Python APIs for trawler',
    'author': 'Alex Sparrow',
    'author_email': 'alex@alexsparrow.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from datetime import datetime, date
from typing import Optional
from pprint import pprint
import click
from trawler.sql.extract import extract_sql
from trawler.agent.jobs import run_scheduler
import logging

logging.basicConfig()

@click.group()
@click.option('--debug', default=False, is_flag=True)
def main(debug: bool):
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)

@main.command()
@click.argument("uri")
@click.option("--override-dbname")
def sql(uri: str, override_dbname: Optional[str]):
    pprint(extract_sql(uri, override_dbname))

@main.command()
@click.argument("config")
@click.option("--now", is_flag=True)
def run(config: str, now: bool):
    run_scheduler(config, now)

if __name__ == "__main__":
    main()



from urllib.parse import urlparse
from sqlalchemy.inspection import inspect
from sqlalchemy.engine import create_engine
from typing import Optional

from trawler.schema import Table, Field, Relation
from trawler.graph import Graph
from trawler.sql.metrics import get_column_metrics, get_table_metrics
import logging
import os

LOG = logging.getLogger(__name__)

def database_urn(scheme, database):
    return f"urn:tr:sql-table::{scheme}/{database}"


def table_urn(scheme, database, schema, table):
    return f"urn:tr:sql-table::{scheme}/{database}/{schema}/{table}"


def column_urn(scheme, database, schema, table, field):
    return f"urn:tr:sql-column::{scheme}/{database}/{schema}/{table}/{field}"


def constraint_urn(scheme, database, schema, table, relation):
    return (
        f"urn:tr:sql-constraint::{scheme}/{database}/{schema}/{table}/{relation}"
    )


def inspect_table(inspector, schema, table_name):
    comments = inspector.get_table_comment(table_name, schema=schema)["text"]

    table = Table(table_name, [], [], comments)
    primary_key_columns = set(
        inspector.get_pk_constraint(table_name, schema=schema)["constrained_columns"]
    )
    for column in inspector.get_columns(table_name, schema=schema):
        LOG.debug(type(column["type"]))
        table.fields.append(
            Field(
                name=column["name"],
                field_type=column["type"],
                comment=column["comment"],
                nullable=column["nullable"],
                is_primary_key=column["name"] in primary_key_columns,
            )
        )
    for fkey in inspector.get_foreign_keys(table_name, schema=schema):
        table.relations.append(
            Relation(
                fkey["name"],
                fkey["constrained_columns"],
                fkey["referred_table"],
                fkey["referred_columns"],
            )
        )

    return table


def extract_sql(uri: str, override_dbname: Optional[str] = None, project=None):
    """Capture a schema from a running database via sqlalchemy"""
    engine = create_engine(uri)
    inspector = inspect(engine)

    parsed_uri = urlparse(uri)
    host = parsed_uri.netloc.split("@")[-1]
    database = parsed_uri.path.strip("/")
    scheme = parsed_uri.scheme

    db_name = override_dbname or f"{host}-{database}"

    g = Graph()

    tables = []
    constraints = []

    LOG.info(f"starting capture: {db_name}")
    for schema in inspector.get_schema_names():
        for table_name in inspector.get_table_names(schema):
            LOG.info(f"catpure: {schema}/{table_name}")
            table = inspect_table(inspector, schema, table_name)

            sql_table = g.SqlTable(
                table_urn(scheme, db_name, schema, table_name),
                name=table_name,
                tr__hasField=[
                    g.SqlColumn(
                        column_urn(
                            scheme,
                            db_name,
                            schema,
                            table_name,
                            field.name,
                        ),
                        name=field.name,
                        tr__dataType=field.field_type.__visit_name__,
                        tr__isNullable=field.nullable,
                        tr__comment=field.comment,
                        tr__hasConstraint=[
                            constraint_urn(
                                scheme,
                                db_name,
                                schema,
                                table_name,
                                relation.name,
                            )
                            for relation in table.relations
                            if field.name in relation.source_fields
                        ],
                        **get_column_metrics(
                            engine, field.field_type, f'{schema}."{table_name}"', f'"{field.name}"'
                        ),
                    )
                    for field in table.fields
                ],
                **get_table_metrics(engine, f'{schema}."{table_name}"'),
            )
            tables.append(sql_table)

            for relation in table.relations:
                constraints.append(
                    g.SqlConstraint(
                        constraint_urn(
                            scheme, db_name, schema, table_name, relation.name
                        ),
                        name=relation.name,
                        tr__constrains=[
                            {
                                "@id": column_urn(
                                    scheme,
                                    db_name,
                                    schema,
                                    relation.target_object,
                                    field,
                                ),
                            }
                            for field in relation.target_fields
                        ],
                    )
                )

    db = g.SqlDatabase(
        database_urn(scheme, db_name),
        name=database,
        tr__has=tables + constraints,
    )
    LOG.info(f"Finished capture: {db_name}")

    g = Graph()
    g.add(db)
    return g.store(project=project)
